"""Classes to represent cycle state $S$ as $S_n = S_{0} + \\sum_{i=1}^n \\Delta S_{i}"""
from __future__ import annotations

import dataclasses
import inspect
from collections import UserDict
from dataclasses import dataclass, fields, replace
from functools import singledispatch, wraps
from typing import Generic, List, TypeVar

import numpy as np
import pandas as pd

S = TypeVar("S")
T = TypeVar("T")


@dataclass(frozen=True)
class State:
    """
    Base object for dataclasses which use the Delta mechanism.

    Examples:
        >>> from dataclasses import dataclass, field

        We define a dataclass where each field (which is going to be delta-ed) has additional
        metadata "delta" which describes its delta behaviour.
        >>> @dataclass(frozen=True)
        ... class ListState(State):
        ...    l: List = field(default_factory=list, metadata={"delta": "extend"})
        ...    m: List = field(default_factory=list, metadata={"delta": "replace"})

        Now we instantiate the dataclass...
        >>> l = ListState(l=list("abc"), m=list("xyz"))
        >>> l
        ListState(l=['a', 'b', 'c'], m=['x', 'y', 'z'])

        ... and can add deltas to it. `l` will be extended:
        >>> l + Delta(l=list("def"))
        ListState(l=['a', 'b', 'c', 'd', 'e', 'f'], m=['x', 'y', 'z'])

        ... wheras `m` will be replaced:
        >>> l + Delta(m=list("uvw"))
        ListState(l=['a', 'b', 'c'], m=['u', 'v', 'w'])

        ... they can be chained:
        >>> l + Delta(l=list("def")) + Delta(m=list("uvw"))
        ListState(l=['a', 'b', 'c', 'd', 'e', 'f'], m=['u', 'v', 'w'])

        ... and we update multiple fields with one Delta:
        >>> l + Delta(l=list("ghi"), m=list("rst"))
        ListState(l=['a', 'b', 'c', 'g', 'h', 'i'], m=['r', 's', 't'])

        Passing a nonexistent field will cause an error:
        >>> l + Delta(o="not a field")
        Traceback (most recent call last):
        ...
        AttributeError: key=`o` is missing on ListState(l=['a', 'b', 'c'], m=['x', 'y', 'z'])

        We can also use the `.update` method to do the same thing:
        >>> l.update(l=list("ghi"), m=list("rst"))
        ListState(l=['a', 'b', 'c', 'g', 'h', 'i'], m=['r', 's', 't'])

        We can also define fields which `append` the last result:
        >>> @dataclass(frozen=True)
        ... class AppendState(State):
        ...    n: List = field(default_factory=list, metadata={"delta": "append"})

        >>> m = AppendState(n=list("ɑβɣ"))
        >>> m
        AppendState(n=['ɑ', 'β', 'ɣ'])

        `n` will be appended:
        >>> m + Delta(n="∂")
        AppendState(n=['ɑ', 'β', 'ɣ', '∂'])

    """

    def __add__(self, other: Delta):
        updates = dict()
        for key, other_value in other.data.items():
            try:
                self_field = next(filter(lambda f: f.name == key, fields(self)))
            except StopIteration:
                raise AttributeError("key=`%s` is missing on %s" % (key, self))
            delta_behavior = self_field.metadata["delta"]
            self_value = getattr(self, key)
            if delta_behavior == "extend":
                extended_value = extend(self_value, other_value)
                updates[key] = extended_value
            elif delta_behavior == "append":
                appended_value = append(self_value, other_value)
                updates[key] = appended_value
            elif delta_behavior == "replace":
                updates[key] = other_value
            else:
                raise NotImplementedError(
                    "delta_behaviour=`%s` not implemented" % (delta_behavior)
                )

        new = replace(self, **updates)
        return new

    def update(self, **kwargs):
        return self + Delta(**kwargs)


class Delta(UserDict, Generic[S]):
    """
    Represents a delta where the base object determines the extension behavior.

    Examples:
        >>> from dataclasses import dataclass

        First we define the dataclass to act as the basis:
        >>> from typing import Optional, List
        >>> @dataclass(frozen=True)
        ... class ListState:
        ...     l: Optional[List] = None
        ...     m: Optional[List] = None
        ...
    """

    pass


Result = Delta
"""`Result` is an alias for `Delta`."""


@singledispatch
def extend(a, b):
    """
    Function to extend supported datatypes.

    """
    raise NotImplementedError("`extend` not implemented for %s, %s" % (a, b))


@extend.register(list)
def extend_list(a, b):
    """
    Examples:
        >>> extend([], [])
        []

        >>> extend([1,2], [3])
        [1, 2, 3]
    """
    return a + b


@extend.register(pd.DataFrame)
def extend_pd_dataframe(a, b):
    """
    Examples:
        >>> extend(pd.DataFrame({"a": []}), pd.DataFrame({"a": []}))
        Empty DataFrame
        Columns: [a]
        Index: []

        >>> extend(pd.DataFrame({"a": [1,2,3]}), pd.DataFrame({"a": [4,5,6]}))
           a
        0  1
        1  2
        2  3
        3  4
        4  5
        5  6
    """
    return pd.concat((a, b), ignore_index=True)


def append(a: List[T], b: T) -> List[T]:
    # TODO: add DOCTESTS
    return a + [b]


@extend.register(np.ndarray)
def extend_np_ndarray(a, b):
    """
    Examples:
        >>> extend(np.array([(1,2,3), (4,5,6)]), np.array([(7,8,9)]))
        array([[1, 2, 3],
               [4, 5, 6],
               [7, 8, 9]])
    """
    return np.row_stack([a, b])


@extend.register(dict)
def extend_dict(a, b):
    """
    Examples:
        >>> extend({"a": "cats"}, {"b": "dogs"})
        {'a': 'cats', 'b': 'dogs'}
    """
    return dict(a, **b)


def wrap_to_use_state(f):
    """Decorator to make target `f` into a function on a `State` and `**kwargs`.

    This wrapper makes it easier to pass arguments to a function from a State.

    It was inspired by the pytest "fixtures" mechanism.

    Args:
        f:

    Returns:

    Examples:
        >>> from autora.state.delta import State, Delta
        >>> from dataclasses import dataclass, field
        >>> import pandas as pd
        >>> from typing import List, Optional

        The `State` it operates on needs to have the metadata described in the state module:
        >>> @dataclass(frozen=True)
        ... class S(State):
        ...     conditions: List[int] = field(metadata={"delta": "replace"})

        We indicate the inputs required by the parameter names.
        The output must be a `Delta` object.
        >>> from autora.state.delta import Delta
        >>> @wrap_to_use_state
        ... def experimentalist(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return Delta(conditions=new_conditions)

        >>> experimentalist(S(conditions=[1,2,3,4]))
        S(conditions=[11, 12, 13, 14])

        >>> experimentalist(S(conditions=[101,102,103,104]))
        S(conditions=[111, 112, 113, 114])

        >>> from autora.variable import VariableCollection, Variable
        >>> from sklearn.base import BaseEstimator
        >>> from sklearn.linear_model import LinearRegression

        >>> @wrap_to_use_state
        ... def theorist(experiment_data: pd.DataFrame, variables: VariableCollection, **kwargs):
        ...     ivs = [v.name for v in variables.independent_variables]
        ...     dvs = [v.name for v in variables.dependent_variables]
        ...     X, y = experiment_data[ivs], experiment_data[dvs]
        ...     new_model = LinearRegression(fit_intercept=True).set_params(**kwargs).fit(X, y)
        ...     return Delta(model=new_model)

        >>> @dataclass(frozen=True)
        ... class T(State):
        ...     variables: VariableCollection  # field(metadata={"delta":... }) omitted ∴ immutable
        ...     experiment_data: pd.DataFrame = field(metadata={"delta": "extend"})
        ...     model: Optional[BaseEstimator] = field(metadata={"delta": "replace"}, default=None)

        >>> t = T(
        ...     variables=VariableCollection(independent_variables=[Variable("x")],
        ...                                  dependent_variables=[Variable("y")]),
        ...     experiment_data=pd.DataFrame({"x": [0,1,2,3,4], "y": [2,3,4,5,6]})
        ... )
        >>> t_prime = theorist(t)
        >>> t_prime.model.coef_, t_prime.model.intercept_
        (array([[1.]]), array([2.]))

        Arguments from the state can be overridden by passing them in as keyword arguments (kwargs):
        >>> theorist(t, experiment_data=pd.DataFrame({"x": [0,1,2,3], "y": [12,13,14,15]}))\\
        ...     .model.intercept_
        array([12.])

        ... and other arguments supported by the inner function can also be passed
        (if and only if the inner function allows for and handles `**kwargs` arguments alongside
        the values from the state).
        >>> theorist(t, fit_intercept=False).model.intercept_
        0.0

        Any parameters not provided by the state must be provided by default values or by the
        caller. If the default is specified:
        >>> @wrap_to_use_state
        ... def experimentalist(conditions, offset=25):
        ...     new_conditions = [c + offset for c in conditions]
        ...     return Delta(conditions=new_conditions)

        ... then it need not be passed.
        >>> experimentalist(S(conditions=[1,2,3,4]))
        S(conditions=[26, 27, 28, 29])

        If a default isn't specified:
        >>> @wrap_to_use_state
        ... def experimentalist(conditions, offset):
        ...     new_conditions = [c + offset for c in conditions]
        ...     return Delta(conditions=new_conditions)

        ... then calling the experimentalist without it will throw an error:
        >>> experimentalist(S(conditions=[1,2,3,4]))
        Traceback (most recent call last):
        ...
        TypeError: experimentalist() missing 1 required positional argument: 'offset'

        ... which can be fixed by passing the argument as a keyword to the wrapped function.
        >>> experimentalist(S(conditions=[1,2,3,4]), offset=2)
        S(conditions=[3, 4, 5, 6])

    """
    # Get the set of parameter names from function f's signature
    parameters_ = set(inspect.signature(f).parameters.keys())

    @wraps(f)
    def _f(state_: S, /, **kwargs) -> S:
        # Get the parameters needed which are available from the state_.
        # All others must be provided as kwargs or default values on f.
        assert dataclasses.is_dataclass(state_)
        from_state = parameters_.intersection(
            {i.name for i in dataclasses.fields(state_)}
        )
        arguments_from_state = {k: getattr(state_, k) for k in from_state}
        arguments = dict(arguments_from_state, **kwargs)
        delta = f(**arguments)
        new_state = state_ + delta
        return new_state

    return _f



import random
from typing import Iterable, Sequence, Union

from autora.utils.deprecation import deprecated_alias


def random_sample(conditions: Union[Iterable, Sequence], num_samples: int = 1):
    """
    Uniform random sampling without replacement from a pool of conditions.
    Args:
        conditions: Pool of conditions
        n: number of samples to collect

    Returns: Sampled pool

    """

    if isinstance(conditions, Iterable):
        conditions = list(conditions)
    random.shuffle(conditions)
    samples = conditions[0:num_samples]

    return samples


random_sampler = deprecated_alias(random_sample, "random_sampler")



from enum import Enum
from typing import (
    Any,
    Dict,
    Generic,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Set,
    TypeVar,
    Union,
    runtime_checkable,
)

from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator

from autora.variable import VariableCollection

State = TypeVar("State")


class ResultKind(str, Enum):
    """
    Kinds of results which can be held in the Result object.

    Examples:
        >>> ResultKind.CONDITION is ResultKind.CONDITION
        True

        >>> ResultKind.CONDITION is ResultKind.VARIABLES
        False

        >>> ResultKind.CONDITION == "CONDITION"
        True

        >>> ResultKind.CONDITION == "VARIABLES"
        False

        >>> ResultKind.CONDITION in {ResultKind.CONDITION, ResultKind.PARAMS}
        True

        >>> ResultKind.VARIABLES in {ResultKind.CONDITION, ResultKind.PARAMS}
        False
    """

    CONDITION = "CONDITION"
    OBSERVATION = "OBSERVATION"
    MODEL = "MODEL"
    PARAMS = "PARAMS"
    VARIABLES = "VARIABLES"

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f"{cls_name}.{self.name}"


class SupportsDataKind(Protocol):
    """Object with attributes for `data` and `kind`"""

    data: Optional[Any]
    kind: Optional[ResultKind]


class SupportsStateParamsField(Protocol):
    """Support a state with a params property."""

    params: Dict


class SupportsStateParamsProperty(Protocol):
    """Support a state with a params property."""

    @property
    def params(self) -> Dict:
        ...


SupportsStateParams = Union[SupportsStateParamsField, SupportsStateParamsProperty]


class SupportsControllerStateFields(Protocol):
    """Support representing snapshots of a controller state as mutable fields."""

    variables: VariableCollection
    params: Dict
    conditions: Sequence[ArrayLike]
    observations: Sequence[ArrayLike]
    models: Sequence[BaseEstimator]

    def update(self: State, **kwargs) -> State:
        ...


class SupportsControllerStateProperties(Protocol):
    """Support representing snapshots of a controller state as immutable properties."""

    def update(self: State, **kwargs) -> State:
        ...

    @property
    def variables(self) -> VariableCollection:
        ...

    @property
    def params(self) -> Dict:
        ...

    @property
    def conditions(self) -> Sequence[ArrayLike]:
        ...

    @property
    def observations(self) -> Sequence[ArrayLike]:
        ...

    @property
    def models(self) -> Sequence[BaseEstimator]:
        ...


SupportsControllerState = Union[
    SupportsControllerStateFields, SupportsControllerStateProperties
]


class SupportsControllerStateHistory(SupportsControllerStateProperties, Protocol):
    """Represents controller state as a linear sequence of entries."""

    def __init__(self, history: Sequence[SupportsDataKind]):
        ...

    def filter_by(self: State, kind: Optional[Set[Union[str, ResultKind]]]) -> State:
        ...

    @property
    def history(self) -> Sequence[SupportsDataKind]:
        ...


class Executor(Protocol, Generic[State]):
    """A Callable which, given some state, and some parameters, returns an updated state."""

    def __call__(self, __state: State, params: Dict) -> State:
        ...


ExecutorCollection = Mapping[str, Executor]


@runtime_checkable
class SupportsLoadDump(Protocol):
    def dump(self, data, file) -> None:
        ...

    def load(self, file) -> Any:
        ...



""" Classes for storing and passing a cycle's state as an immutable snapshot. """
from dataclasses import dataclass, field
from typing import Dict, List

from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator

from autora.state.delta import Delta
from autora.state.protocol import SupportsControllerStateFields
from autora.variable import VariableCollection


@dataclass(frozen=True)
class Snapshot(SupportsControllerStateFields):
    """An object passed between and updated by processing steps in the Controller."""

    # Single values
    variables: VariableCollection = field(default_factory=VariableCollection)
    params: Dict = field(default_factory=dict)

    # Sequences
    conditions: List[ArrayLike] = field(default_factory=list)
    observations: List[ArrayLike] = field(default_factory=list)
    models: List[BaseEstimator] = field(default_factory=list)

    def update(
        self,
        variables=None,
        params=None,
        conditions=None,
        observations=None,
        models=None,
    ):
        """
        Create a new object with updated values.

        Examples:
            The initial object is empty:
            >>> s0 = Snapshot()
            >>> s0  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(variables=VariableCollection(...), params={}, conditions=[],
                            observations=[], models=[])

            We can update the params using the `.update` method:
            >>> s0.update(params={'first': 'params'})  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(..., params={'first': 'params'}, ...)

            ... but the original object is unchanged:
            >>> s0  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(..., params={}, ...)

            For params, only one object is returned from the respective property:
            >>> s0.update(params={'first': 'params'}).update(params={'second': 'params'}).params
            {'second': 'params'}

            ... and the same applies to variables:
            >>> from autora.variable import VariableCollection, IV
            >>> (s0.update(variables=VariableCollection([IV("1st IV")]))
            ...    .update(variables=VariableCollection([IV("2nd IV")]))).variables
            VariableCollection(independent_variables=[IV(name='2nd IV',...)], ...)

            When we update the conditions, observations or models, the respective list is extended:
            >>> s3 = s0.update(models=["1st model"])
            >>> s3
            Snapshot(..., models=['1st model'])

            ... so we can see the history of all the models, for instance.
            >>> s3.update(models=["2nd model"])
            Snapshot(..., models=['1st model', '2nd model'])

            The same applies to observations:
            >>> s4 = s0.update(observations=["1st observation"])
            >>> s4
            Snapshot(..., observations=['1st observation'], ...)

            >>> s4.update(observations=["2nd observation"])  # doctest: +ELLIPSIS
            Snapshot(..., observations=['1st observation', '2nd observation'], ...)


            The same applies to conditions:
            >>> s5 = s0.update(conditions=["1st condition"])
            >>> s5
            Snapshot(..., conditions=['1st condition'], ...)

            >>> s5.update(conditions=["2nd condition"])  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(..., conditions=['1st condition', '2nd condition'], ...)

            You can also update with multiple conditions, observations and models:
            >>> s0.update(conditions=['c1', 'c2'])
            Snapshot(..., conditions=['c1', 'c2'], ...)

            >>> s0.update(models=['m1', 'm2'], variables={'m': 1})
            Snapshot(variables={'m': 1}, ..., models=['m1', 'm2'])

            >>> s0.update(models=['m1'], observations=['o1'], variables={'m': 1})
            Snapshot(variables={'m': 1}, ..., observations=['o1'], models=['m1'])


            Inputs to models, observations and conditions must be Lists
            which can be cast to lists:
            >>> s0.update(models='m1')  # doctest: +ELLIPSIS
            Traceback (most recent call last):
            ...
            AssertionError: 'm1' must be a list, e.g. `['m1']`?)

        """

        def _coalesce_lists(old, new):
            assert isinstance(
                old, List
            ), f"{repr(old)} must be a list, e.g. `[{repr(old)}]`?)"
            if new is not None:
                assert isinstance(
                    new, List
                ), f"{repr(new)} must be a list, e.g. `[{repr(new)}]`?)"
                return old + list(new)
            else:
                return old

        variables_ = variables or self.variables
        params_ = params or self.params
        conditions_ = _coalesce_lists(self.conditions, conditions)
        observations_ = _coalesce_lists(self.observations, observations)
        models_ = _coalesce_lists(self.models, models)
        return Snapshot(variables_, params_, conditions_, observations_, models_)

    def __add__(self, other: Delta):
        """
        Add a delta to the object.

        Examples:
            The initial object is empty:
            >>> s0 = Snapshot()
            >>> s0  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(variables=VariableCollection(...), params={}, conditions=[],
                            observations=[], models=[])

            We can update the params using the `+` operator:
            >>> from autora.state.delta import Delta
            >>> s0 + Delta(params={'first': 'params'})  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(..., params={'first': 'params'}, ...)

            ... but the original object is unchanged:
            >>> s0  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(..., params={}, ...)

            For params, only one object is returned from the respective property:
            >>> (s0 + Delta(params={'first': 'params'}) + Delta(params={'second':'params'})).params
            {'second': 'params'}

            ... and the same applies to variables:
            >>> from autora.variable import VariableCollection, IV
            >>> (s0 + Delta(variables=VariableCollection([IV("1st IV")])) +
            ...    Delta(variables=VariableCollection([IV("2nd IV")]))).variables
            VariableCollection(independent_variables=[IV(name='2nd IV',...)], ...)

            When we update the conditions, observations or models, the respective list is extended:
            >>> s3 = s0 + Delta(models=["1st model"])
            >>> s3
            Snapshot(..., models=['1st model'])

            ... so we can see the history of all the models, for instance.
            >>> s3 + Delta(models=["2nd model"])
            Snapshot(..., models=['1st model', '2nd model'])

            The same applies to observations:
            >>> s4 = s0 + Delta(observations=["1st observation"])
            >>> s4
            Snapshot(..., observations=['1st observation'], ...)

            >>> s4 + Delta(observations=["2nd observation"])  # doctest: +ELLIPSIS
            Snapshot(..., observations=['1st observation', '2nd observation'], ...)


            The same applies to conditions:
            >>> s5 = s0 + Delta(conditions=["1st condition"])
            >>> s5
            Snapshot(..., conditions=['1st condition'], ...)

            >>> s5 + Delta(conditions=["2nd condition"])  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            Snapshot(..., conditions=['1st condition', '2nd condition'], ...)

            You can also update with multiple conditions, observations and models:
            >>> s0 + Delta(conditions=['c1', 'c2'])
            Snapshot(..., conditions=['c1', 'c2'], ...)

            >>> s0 + Delta(models=['m1', 'm2'], variables={'m': 1})
            Snapshot(variables={'m': 1}, ..., models=['m1', 'm2'])

            >>> s0 + Delta(models=['m1'], observations=['o1'], variables={'m': 1})
            Snapshot(variables={'m': 1}, ..., observations=['o1'], models=['m1'])


            Inputs to models, observations and conditions must be Lists
            which can be cast to lists:
            >>> s0 + Delta(models='m1')  # doctest: +ELLIPSIS
            Traceback (most recent call last):
            ...
            AssertionError: 'm1' must be a list, e.g. `['m1']`?)
        """
        return self.update(**other)



"""Utilities to wrap common theorist, experimentalist and experiment runners as `f(State)`.
so that $n$ processes $f_i$ on states $S$ can be represented as
$$f_n(...(f_1(f_0(S))))$$
"""
from __future__ import annotations

from typing import Callable, Iterable, TypeVar

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator

from autora.experimentalist.pipeline import Pipeline
from autora.state.delta import Delta, State, wrap_to_use_state
from autora.variable import VariableCollection

S = TypeVar("S")
X = TypeVar("X")
Y = TypeVar("Y")
XY = TypeVar("XY")
Executor = Callable[[State], State]


def theorist_from_estimator(estimator: BaseEstimator) -> Executor:
    """
    Convert a scikit-learn compatible estimator into a function on a `State` object.

    Supports passing additional `**kwargs` which are used to update the estimator's params
    before fitting.
    """

    @wrap_to_use_state
    def theorist(
        experiment_data: pd.DataFrame, variables: VariableCollection, **kwargs
    ):
        ivs = [v.name for v in variables.independent_variables]
        dvs = [v.name for v in variables.dependent_variables]
        X, y = experiment_data[ivs], experiment_data[dvs]
        new_model = estimator.set_params(**kwargs).fit(X, y)
        return Delta(model=new_model)

    return theorist


def experiment_runner_from_x_to_y_function(f: Callable[[X], Y]) -> Executor:
    """Wrapper for experiment_runner of the form $f(x) \rarrow y$, where `f` returns just the $y$
    values"""

    @wrap_to_use_state
    def experiment_runner(conditions: pd.DataFrame, **kwargs):
        x = conditions
        y = f(x, **kwargs)
        experiment_data = pd.DataFrame.merge(x, y, left_index=True, right_index=True)
        return Delta(experiment_data=experiment_data)

    return experiment_runner


def experiment_runner_from_x_to_xy_function(f: Callable[[X], XY]) -> Executor:
    """Wrapper for experiment_runner of the form $f(x) \rarrow (x,y)$, where `f`
    returns both $x$ and $y$ values in a complete dataframe."""

    @wrap_to_use_state
    def experiment_runner(conditions: pd.DataFrame, **kwargs):
        x = conditions
        experiment_data = f(x, **kwargs)
        return Delta(experiment_data=experiment_data)

    return experiment_runner


def experimentalist_from_pipeline(pipeline: Pipeline) -> Executor:
    """Wrapper for experimentalists of the form $f() \rarrow x$, where `f`
    returns both $x$ and $y$ values in a complete dataframe."""

    @wrap_to_use_state
    def experimentalist(params):
        conditions = pipeline(**params)
        if isinstance(conditions, (pd.DataFrame, np.ndarray, np.recarray)):
            conditions_ = conditions
        elif isinstance(conditions, Iterable):
            conditions_ = np.array(list(conditions))
        else:
            raise NotImplementedError("type `%s` is not supported" % (type(conditions)))
        return Delta(conditions=conditions_)

    return experimentalist



from itertools import product
from typing import List

from autora.variable import IV


def grid_pool(ivs: List[IV]):
    """Creates exhaustive pool from discrete values using a Cartesian product of sets"""
    # Get allowed values for each IV
    l_iv_values = []
    for iv in ivs:
        assert iv.allowed_values is not None, (
            f"gridsearch_pool only supports independent variables with discrete allowed values, "
            f"but allowed_values is None on {iv=} "
        )
        l_iv_values.append(iv.allowed_values)

    # Return Cartesian product of all IV values
    return product(*l_iv_values)



""" Functions for handling cycle-state-dependent parameters. """
from __future__ import annotations

import copy
import logging
from typing import Dict, Mapping

import numpy as np

from autora.state.protocol import SupportsControllerState
from autora.utils.deprecation import deprecate as deprecate
from autora.utils.dictionary import LazyDict

_logger = logging.getLogger(__name__)


def _get_state_dependent_properties(state: SupportsControllerState):
    """
    Examples:
        Even with an empty data object, we can initialize the dictionary,
        >>> from autora.state.snapshot import Snapshot
        >>> state_dependent_properties = _get_state_dependent_properties(Snapshot())

        ... but it will raise an exception if a value isn't yet available when we try to use it
        >>> state_dependent_properties["%models[-1]%"] # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        IndexError: list index out of range

        Nevertheless, we can iterate through its keys no problem:
        >>> [key for key in state_dependent_properties.keys()] # doctest: +NORMALIZE_WHITESPACE
        ['%observations.ivs[-1]%', '%observations.dvs[-1]%', '%observations.ivs%',
         '%observations.dvs%', '%experiment_data.conditions[-1]%',
         '%experiment_data.observations[-1]%', '%experiment_data.conditions%',
         '%experiment_data.observations%', '%models[-1]%', '%models%']

    """

    n_ivs = len(state.variables.independent_variables)
    n_dvs = len(state.variables.dependent_variables)
    state_dependent_property_dict = LazyDict(
        {
            "%observations.ivs[-1]%": deprecate(
                lambda: np.array(state.observations[-1])[:, 0:n_ivs],
                "%observations.ivs[-1]% is deprecated, "
                "use %experiment_data.conditions[-1]% instead.",
            ),
            "%observations.dvs[-1]%": deprecate(
                lambda: np.array(state.observations[-1])[:, n_ivs:],
                "%observations.dvs[-1]% is deprecated, "
                "use %experiment_data.observations[-1]% instead.",
            ),
            "%observations.ivs%": deprecate(
                lambda: np.row_stack(
                    [np.empty([0, n_ivs + n_dvs])] + list(state.observations)
                )[:, 0:n_ivs],
                "%observations.ivs% is deprecated, use %experiment_data.conditions% instead.",
            ),
            "%observations.dvs%": deprecate(
                lambda: np.row_stack(state.observations)[:, n_ivs:],
                "%observations.dvs% is deprecated, "
                "use %experiment_data.observations% instead",
            ),
            "%experiment_data.conditions[-1]%": lambda: np.array(
                state.observations[-1]
            )[:, 0:n_ivs],
            "%experiment_data.observations[-1]%": lambda: np.array(
                state.observations[-1]
            )[:, n_ivs:],
            "%experiment_data.conditions%": lambda: np.row_stack(
                [np.empty([0, n_ivs + n_dvs])] + list(state.observations)
            )[:, 0:n_ivs],
            "%experiment_data.observations%": lambda: np.row_stack(state.observations)[
                :, n_ivs:
            ],
            "%models[-1]%": lambda: state.models[-1],
            "%models%": lambda: state.models,
        }
    )
    return state_dependent_property_dict


def _resolve_properties(params: Dict, state_dependent_properties: Mapping):
    """
    Resolve state-dependent properties inside a nested dictionary.

    In this context, a state-dependent-property is a string which is meant to be replaced by its
    updated, current value before the dictionary is used. A state-dependent property might be
    something like "the last theorist available" or "all the experimental results until now".

    Args:
        params: a (nested) dictionary of keys and values, where some values might be
            "cycle property names"
        state_dependent_properties: a dictionary of "property names" and their "real values"

    Returns: a (nested) dictionary where "property names" are replaced by the "real values"

    Examples:

        >>> params_0 = {"key": "%foo%"}
        >>> cycle_properties_0 = {"%foo%": 180}
        >>> _resolve_properties(params_0,cycle_properties_0)
        {'key': 180}

        >>> params_1 = {"key": "%bar%", "nested_dict": {"inner_key": "%foobar%"}}
        >>> cycle_properties_1 = {"%bar%": 1, "%foobar%": 2}
        >>> _resolve_properties(params_1,cycle_properties_1)
        {'key': 1, 'nested_dict': {'inner_key': 2}}

        >>> params_2 = {"key": "baz"}
        >>> _resolve_properties(params_2,cycle_properties_1)
        {'key': 'baz'}

    """
    params_ = copy.copy(params)
    for key, value in params_.items():
        if isinstance(value, dict):
            params_[key] = _resolve_properties(value, state_dependent_properties)
        elif isinstance(value, str) and (
            value in state_dependent_properties
        ):  # value is a key in the cycle_properties dictionary
            params_[key] = state_dependent_properties[value]
        else:
            _logger.debug(f"leaving {params=} unchanged")

    return params_


def resolve_state_params(params: Dict, state: SupportsControllerState) -> Dict:
    """
    Returns the `params` attribute of the input, with `cycle properties` resolved.

    Examples:
        >>> from autora.state.history import History
        >>> params = {"experimentalist": {"source": "%models[-1]%"}}
        >>> s = History(models=["the first model", "the second model"])
        >>> resolve_state_params(params, s)
        {'experimentalist': {'source': 'the second model'}}

    """
    state_dependent_properties = _get_state_dependent_properties(state)
    resolved_params = _resolve_properties(params, state_dependent_properties)
    return resolved_params



from functools import partial

import numpy as np
import pytest

from autora.experimentalist.pipeline import make_pipeline
from autora.experimentalist.pooler.grid import grid_pool
from autora.experimentalist.pooler.random_pooler import random_pool
from autora.experimentalist.sampler.random_sampler import random_sample
from autora.variable import DV, IV, ValueType, VariableCollection


def weber_filter(values):
    return filter(lambda s: s[0] <= s[1], values)


def test_random_pooler_experimentalist(metadata):
    """
    Tests the implementation of a random pooler.
    """
    num_samples = 10

    conditions = random_pool(metadata.independent_variables, num_samples=num_samples)

    conditions = np.array(list(conditions))

    assert conditions.shape[0] == num_samples
    assert conditions.shape[1] == len(metadata.independent_variables)
    for condition in conditions:
        for idx, value in enumerate(condition):
            assert value in metadata.independent_variables[idx].allowed_values


def test_random_sampler_experimentalist(metadata):
    """
    Tests the implementation of the experimentalist pipeline with an exhaustive pool of discrete
    values, Weber filter, random selector. Tests two different implementations of the pool function
    as a callable and passing in as interator/generator.

    """

    n_trials = 25  # Number of trails for sampler to select

    # ---Implementation 1 - Pool using Callable via partial function----
    # Set up pipeline functions with partial
    pooler_callable = partial(grid_pool, ivs=metadata.independent_variables)
    sampler = partial(random_sample, num_samples=n_trials)
    pipeline_random_samp = make_pipeline(
        [pooler_callable, weber_filter, sampler],
    )

    results = pipeline_random_samp.run()

    # ***Checks***
    # Gridsearch pool is working as expected
    _, pool = pipeline_random_samp.steps[0]
    pool_len = len(list(pool()))
    pool_len_expected = np.prod(
        [len(s.allowed_values) for s in metadata.independent_variables]
    )
    assert pool_len == pool_len_expected

    # Is sampling the number of trials we expect
    assert len(results) == n_trials

    # Filter is selecting where IV1 >= IV2
    assert all([s[0] <= s[1] for s in results])

    # Is sampling randomly. Runs 10 times and checks if consecutive runs are equal.
    # Assert will fail if all 9 pairs return equal.
    l_results = [pipeline_random_samp.run() for s in range(10)]
    assert not np.all(
        [
            np.array_equal(l_results[i], l_results[i + 1])
            for i, s in enumerate(l_results)
            if i < len(l_results) - 1
        ]
    )


def test_random_experimentalist_generator(metadata):
    n_trials = 25  # Number of trails for sampler to select

    pooler_generator = grid_pool(metadata.independent_variables)
    sampler = partial(random_sample, num_samples=n_trials)
    pipeline_random_samp_poolgen = make_pipeline(
        [pooler_generator, weber_filter, sampler]
    )

    results_poolgen = list(pipeline_random_samp_poolgen.run())

    # Is sampling the number of trials we expect
    assert len(results_poolgen) == n_trials

    # Filter is selecting where IV1 >= IV2
    assert all([s[0] <= s[1] for s in results_poolgen])

    # This will fail
    # The Generator is exhausted after the first run and the pool is not regenerated when pipeline
    # is run again. The pool should be set up as a callable if the pipeline is to be rerun.
    results_poolgen2 = pipeline_random_samp_poolgen.run()
    assert len(results_poolgen2) == 0


@pytest.fixture
def metadata():
    # Specify independent variables
    iv1 = IV(
        name="S1",
        allowed_values=np.linspace(0, 5, 5),
        units="intensity",
        variable_label="Stimulus 1 Intensity",
    )

    iv2 = IV(
        name="S2",
        allowed_values=np.linspace(0, 5, 5),
        units="intensity",
        variable_label="Stimulus 2 Intensity",
    )

    iv3 = IV(
        name="S3",
        allowed_values=[0, 1],
        units="binary",
        variable_label="Stimulus 3 Binary",
    )

    # Specify dependent variable with type
    # The experimentalist pipeline doesn't actually use DVs, they are just specified here for
    # example.
    dv1 = DV(
        name="difference_detected",
        value_range=(0, 1),
        units="probability",
        variable_label="P(difference detected)",
        type=ValueType.SIGMOID,
    )

    dv2 = DV(
        name="difference_detected_sample",
        value_range=(0, 1),
        units="response",
        variable_label="difference detected",
        type=ValueType.PROBABILITY_SAMPLE,
    )
    # Variable collection with ivs and dvs
    metadata = VariableCollection(
        independent_variables=[iv1, iv2, iv3],
        dependent_variables=[dv1, dv2],
    )

    return metadata



""" Classes for storing and passing a cycle's state as an immutable history. """
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Union

from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator

from autora.state.delta import Delta
from autora.state.protocol import (
    ResultKind,
    SupportsControllerStateHistory,
    SupportsDataKind,
)
from autora.state.snapshot import Snapshot
from autora.variable import VariableCollection


class History(SupportsControllerStateHistory):
    """
    An immutable object for tracking the state and history of an AER cycle.
    """

    def __init__(
        self,
        variables: Optional[VariableCollection] = None,
        params: Optional[Dict] = None,
        conditions: Optional[List[ArrayLike]] = None,
        observations: Optional[List[ArrayLike]] = None,
        models: Optional[List[BaseEstimator]] = None,
        history: Optional[Sequence[Result]] = None,
    ):
        """

        Args:
            variables: a single datum to be marked as "variables"
            params: a single datum to be marked as "params"
            conditions: an iterable of data, each to be marked as "conditions"
            observations: an iterable of data, each to be marked as "observations"
            models: an iterable of data, each to be marked as "models"
            history: an iterable of Result objects to be used as the initial history.

        Examples:
            Empty input leads to an empty state:
            >>> History()
            History([])

            ... or with values for any or all of the parameters:
            >>> from autora.variable import VariableCollection
            >>> History(variables=VariableCollection()) # doctest: +ELLIPSIS
            History([Result(data=VariableCollection(...), kind=ResultKind.VARIABLES)])

            >>> History(params={"some": "params"})
            History([Result(data={'some': 'params'}, kind=ResultKind.PARAMS)])

            >>> History(conditions=["a condition"])
            History([Result(data='a condition', kind=ResultKind.CONDITION)])

            >>> History(observations=["an observation"])
            History([Result(data='an observation', kind=ResultKind.OBSERVATION)])

            >>> from sklearn.linear_model import LinearRegression
            >>> History(models=[LinearRegression()])
            History([Result(data=LinearRegression(), kind=ResultKind.MODEL)])

            Parameters passed to the constructor are included in the history in the following order:
            `history`, `variables`, `params`, `conditions`, `observations`, `models`
            >>> History(models=['m1', 'm2'], conditions=['c1', 'c2'],
            ...     observations=['o1', 'o2'], params={'a': 'param'},
            ...     variables=VariableCollection(),
            ...     history=[Result("from history", ResultKind.VARIABLES)]
            ... )  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            History([Result(data='from history', kind=ResultKind.VARIABLES),
                                    Result(data=VariableCollection(...), kind=ResultKind.VARIABLES),
                                    Result(data={'a': 'param'}, kind=ResultKind.PARAMS),
                                    Result(data='c1', kind=ResultKind.CONDITION),
                                    Result(data='c2', kind=ResultKind.CONDITION),
                                    Result(data='o1', kind=ResultKind.OBSERVATION),
                                    Result(data='o2', kind=ResultKind.OBSERVATION),
                                    Result(data='m1', kind=ResultKind.MODEL),
                                    Result(data='m2', kind=ResultKind.MODEL)])
        """
        self.data: List

        if history is not None:
            self.data = list(history)
        else:
            self.data = []

        self.data += _init_result_list(
            variables=variables,
            params=params,
            conditions=conditions,
            observations=observations,
            models=models,
        )

    def update(
        self,
        variables=None,
        params=None,
        conditions=None,
        observations=None,
        models=None,
        history=None,
    ):
        """
        Create a new object with updated values.

        Examples:
            The initial object is empty:
            >>> h0 = History()
            >>> h0
            History([])

            We can update the variables using the `.update` method:
            >>> from autora.variable import VariableCollection
            >>> h1 = h0.update(variables=VariableCollection())
            >>> h1  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            History([Result(data=VariableCollection(...), kind=ResultKind.VARIABLES)])

            ... the original object is unchanged:
            >>> h0
            History([])

            We can update the variables again:
            >>> h2 = h1.update(variables=VariableCollection(["some IV"]))
            >>> h2._by_kind  # doctest: +ELLIPSIS
            Snapshot(variables=VariableCollection(independent_variables=['some IV'],...), ...)

            ... and we see that there is only ever one variables object returned.

            Params is treated the same way as variables:
            >>> hp = h0.update(params={'first': 'params'})
            >>> hp
            History([Result(data={'first': 'params'}, kind=ResultKind.PARAMS)])

            ... where only the most recent "params" object is returned from the `.params` property.
            >>> hp = hp.update(params={'second': 'params'})
            >>> hp.params
            {'second': 'params'}

            ... however, the full history of the params objects remains available, if needed:
            >>> hp  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data={'first': 'params'}, kind=ResultKind.PARAMS),
                                    Result(data={'second': 'params'}, kind=ResultKind.PARAMS)])

            When we update the conditions, observations or models, a new entry is added to the
            history:
            >>> h3 = h0.update(models=["1st model"])
            >>> h3  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data='1st model', kind=ResultKind.MODEL)])

            ... so we can see the history of all the models, for instance.
            >>> h3 = h3.update(models=["2nd model"])  # doctest: +NORMALIZE_WHITESPACE
            >>> h3  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data='1st model', kind=ResultKind.MODEL),
                                    Result(data='2nd model', kind=ResultKind.MODEL)])

            ... and the full history of models is available using the `.models` parameter:
            >>> h3.models
            ['1st model', '2nd model']

            The same for the observations:
            >>> h4 = h0.update(observations=["1st observation"])
            >>> h4
            History([Result(data='1st observation', kind=ResultKind.OBSERVATION)])

            >>> h4.update(observations=["2nd observation"]
            ... )  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            History([Result(data='1st observation', kind=ResultKind.OBSERVATION),
                                    Result(data='2nd observation', kind=ResultKind.OBSERVATION)])


            The same for the conditions:
            >>> h5 = h0.update(conditions=["1st condition"])
            >>> h5
            History([Result(data='1st condition', kind=ResultKind.CONDITION)])

            >>> h5.update(conditions=["2nd condition"])  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data='1st condition', kind=ResultKind.CONDITION),
                                    Result(data='2nd condition', kind=ResultKind.CONDITION)])

            You can also update with multiple conditions, observations and models:
            >>> h0.update(conditions=['c1', 'c2'])  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data='c1', kind=ResultKind.CONDITION),
                                    Result(data='c2', kind=ResultKind.CONDITION)])

            >>> h0.update(models=['m1', 'm2'], variables={'m': 1}
            ... ) # doctest: +NORMALIZE_WHITESPACE
            History([Result(data={'m': 1}, kind=ResultKind.VARIABLES),
                     Result(data='m1', kind=ResultKind.MODEL),
                     Result(data='m2', kind=ResultKind.MODEL)])

            >>> h0.update(models=['m1'], observations=['o1'], variables={'m': 1}
            ... )  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data={'m': 1}, kind=ResultKind.VARIABLES),
                     Result(data='o1', kind=ResultKind.OBSERVATION),
                     Result(data='m1', kind=ResultKind.MODEL)])

            We can also update with a complete history:
            >>> History().update(history=[Result(data={'m': 2}, kind=ResultKind.VARIABLES),
            ...                           Result(data='o1', kind=ResultKind.OBSERVATION),
            ...                           Result(data='m1', kind=ResultKind.MODEL)],
            ...                  conditions=['c1']
            ... )  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data={'m': 2}, kind=ResultKind.VARIABLES),
                     Result(data='o1', kind=ResultKind.OBSERVATION),
                     Result(data='m1', kind=ResultKind.MODEL),
                     Result(data='c1', kind=ResultKind.CONDITION)])

        """

        if history is not None:
            history_extension = history
        else:
            history_extension = []

        history_extension += _init_result_list(
            variables=variables,
            params=params,
            conditions=conditions,
            observations=observations,
            models=models,
        )
        new_full_history = self.data + history_extension

        return History(history=new_full_history)

    def __add__(self, other: Delta):
        """The initial object is empty:
        >>> h0 = History()
        >>> h0
        History([])

        We can update the variables using the `.update` method:
        >>> from autora.variable import VariableCollection
        >>> h1 = h0 + Delta(variables=VariableCollection())
        >>> h1  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        History([Result(data=VariableCollection(...), kind=ResultKind.VARIABLES)])

        ... the original object is unchanged:
        >>> h0
        History([])

        We can update the variables again:
        >>> h2 = h1 + Delta(variables=VariableCollection(["some IV"]))
        >>> h2._by_kind  # doctest: +ELLIPSIS
        Snapshot(variables=VariableCollection(independent_variables=['some IV'],...), ...)

        ... and we see that there is only ever one variables object returned.

        Params is treated the same way as variables:
        >>> hp = h0 + Delta(params={'first': 'params'})
        >>> hp
        History([Result(data={'first': 'params'}, kind=ResultKind.PARAMS)])

        ... where only the most recent "params" object is returned from the `.params` property.
        >>> hp = hp + Delta(params={'second': 'params'})
        >>> hp.params
        {'second': 'params'}

        ... however, the full history of the params objects remains available, if needed:
        >>> hp  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data={'first': 'params'}, kind=ResultKind.PARAMS),
                                Result(data={'second': 'params'}, kind=ResultKind.PARAMS)])

        When we update the conditions, observations or models, a new entry is added to the
        history:
        >>> h3 = h0 + Delta(models=["1st model"])
        >>> h3  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data='1st model', kind=ResultKind.MODEL)])

        ... so we can see the history of all the models, for instance.
        >>> h3 = h3 + Delta(models=["2nd model"])  # doctest: +NORMALIZE_WHITESPACE
        >>> h3  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data='1st model', kind=ResultKind.MODEL),
                                Result(data='2nd model', kind=ResultKind.MODEL)])

        ... and the full history of models is available using the `.models` parameter:
        >>> h3.models
        ['1st model', '2nd model']

        The same for the observations:
        >>> h4 = h0 + Delta(observations=["1st observation"])
        >>> h4
        History([Result(data='1st observation', kind=ResultKind.OBSERVATION)])

        >>> h4 + Delta(observations=["2nd observation"]
        ... )  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        History([Result(data='1st observation', kind=ResultKind.OBSERVATION),
                                Result(data='2nd observation', kind=ResultKind.OBSERVATION)])


        The same for the conditions:
        >>> h5 = h0 + Delta(conditions=["1st condition"])
        >>> h5
        History([Result(data='1st condition', kind=ResultKind.CONDITION)])

        >>> h5 + Delta(conditions=["2nd condition"])  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data='1st condition', kind=ResultKind.CONDITION),
                                Result(data='2nd condition', kind=ResultKind.CONDITION)])

        You can also update with multiple conditions, observations and models:
        >>> h0 + Delta(conditions=['c1', 'c2'])  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data='c1', kind=ResultKind.CONDITION),
                                Result(data='c2', kind=ResultKind.CONDITION)])

        >>> h0 + Delta(models=['m1', 'm2'], variables={'m': 1}
        ... ) # doctest: +NORMALIZE_WHITESPACE
        History([Result(data={'m': 1}, kind=ResultKind.VARIABLES),
                 Result(data='m1', kind=ResultKind.MODEL),
                 Result(data='m2', kind=ResultKind.MODEL)])

        >>> h0 + Delta(models=['m1'], observations=['o1'], variables={'m': 1}
        ... )  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data={'m': 1}, kind=ResultKind.VARIABLES),
                 Result(data='o1', kind=ResultKind.OBSERVATION),
                 Result(data='m1', kind=ResultKind.MODEL)])

        We can also update with a complete history:
        >>> History() + Delta(history=[Result(data={'m': 2}, kind=ResultKind.VARIABLES),
        ...                           Result(data='o1', kind=ResultKind.OBSERVATION),
        ...                           Result(data='m1', kind=ResultKind.MODEL)],
        ...                  conditions=['c1']
        ... )  # doctest: +NORMALIZE_WHITESPACE
        History([Result(data={'m': 2}, kind=ResultKind.VARIABLES),
                 Result(data='o1', kind=ResultKind.OBSERVATION),
                 Result(data='m1', kind=ResultKind.MODEL),
                 Result(data='c1', kind=ResultKind.CONDITION)])
        """
        return self.update(**other)

    def __repr__(self):
        return f"{type(self).__name__}({self.history})"

    @property
    def _by_kind(self):
        return _history_to_kind(self.data)

    @property
    def variables(self) -> VariableCollection:
        """

        Examples:
            The initial object is empty:
            >>> h = History()

            ... and returns an emtpy variables object
            >>> h.variables
            VariableCollection(independent_variables=[], dependent_variables=[], covariates=[])

            We can update the variables using the `.update` method:
            >>> from autora.variable import VariableCollection
            >>> h = h.update(variables=VariableCollection(independent_variables=['some IV']))
            >>> h.variables  # doctest: +ELLIPSIS
            VariableCollection(independent_variables=['some IV'], ...)

            We can update the variables again:
            >>> h = h.update(variables=VariableCollection(["some other IV"]))
            >>> h.variables  # doctest: +ELLIPSIS
            VariableCollection(independent_variables=['some other IV'], ...)

            ... and we see that there is only ever one variables object returned."""
        return self._by_kind.variables

    @property
    def params(self) -> Dict:
        """

        Returns:

        Examples:
            Params is treated the same way as variables:
            >>> h = History()
            >>> h = h.update(params={'first': 'params'})
            >>> h.params
            {'first': 'params'}

            ... where only the most recent "params" object is returned from the `.params` property.
            >>> h = h.update(params={'second': 'params'})
            >>> h.params
            {'second': 'params'}

            ... however, the full history of the params objects remains available, if needed:
            >>> h  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data={'first': 'params'}, kind=ResultKind.PARAMS),
                                    Result(data={'second': 'params'}, kind=ResultKind.PARAMS)])
        """
        return self._by_kind.params

    @property
    def conditions(self) -> List[ArrayLike]:
        """
        Returns:

        Examples:
            View the sequence of models with one conditions:
            >>> h = History(conditions=[(1,2,3,)])
            >>> h.conditions
            [(1, 2, 3)]

            ... or more conditions:
            >>> h = h.update(conditions=[(4,5,6),(7,8,9)])  # doctest: +NORMALIZE_WHITESPACE
            >>> h.conditions
            [(1, 2, 3), (4, 5, 6), (7, 8, 9)]

        """
        return self._by_kind.conditions

    @property
    def observations(self) -> List[ArrayLike]:
        """

        Returns:

        Examples:
            The sequence of all observations is returned
            >>> h = History(observations=["1st observation"])
            >>> h.observations
            ['1st observation']

            >>> h = h.update(observations=["2nd observation"])
            >>> h.observations  # doctest: +ELLIPSIS
            ['1st observation', '2nd observation']

        """
        return self._by_kind.observations

    @property
    def models(self) -> List[BaseEstimator]:
        """

        Returns:

        Examples:
            View the sequence of models with one model:
            >>> s = History(models=["1st model"])
            >>> s.models  # doctest: +NORMALIZE_WHITESPACE
            ['1st model']

            ... or more models:
            >>> s = s.update(models=["2nd model"])  # doctest: +NORMALIZE_WHITESPACE
            >>> s.models
            ['1st model', '2nd model']

        """
        return self._by_kind.models

    @property
    def history(self) -> List[Result]:
        """

        Examples:
            We initialze some history:
            >>> h = History(models=['m1', 'm2'], conditions=['c1', 'c2'],
            ...     observations=['o1', 'o2'], params={'a': 'param'},
            ...     variables=VariableCollection(),
            ...     history=[Result("from history", ResultKind.VARIABLES)])

            Parameters passed to the constructor are included in the history in the following order:
            `history`, `variables`, `params`, `conditions`, `observations`, `models`

            >>> h.history  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            [Result(data='from history', kind=ResultKind.VARIABLES),
             Result(data=VariableCollection(...), kind=ResultKind.VARIABLES),
             Result(data={'a': 'param'}, kind=ResultKind.PARAMS),
             Result(data='c1', kind=ResultKind.CONDITION),
             Result(data='c2', kind=ResultKind.CONDITION),
             Result(data='o1', kind=ResultKind.OBSERVATION),
             Result(data='o2', kind=ResultKind.OBSERVATION),
             Result(data='m1', kind=ResultKind.MODEL),
             Result(data='m2', kind=ResultKind.MODEL)]

            If we add a new value, like the params object, the updated value is added to the
            end of the history:
            >>> h = h.update(params={'new': 'param'})
            >>> h.history  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
            [..., Result(data={'new': 'param'}, kind=ResultKind.PARAMS)]

        """
        return self.data

    def filter_by(self, kind: Optional[Set[Union[str, ResultKind]]] = None) -> History:
        """
        Return a copy of the object with only data belonging to the specified kinds.

        Examples:
            >>> h = History(models=['m1', 'm2'], conditions=['c1', 'c2'],
            ...     observations=['o1', 'o2'], params={'a': 'param'},
            ...    variables=VariableCollection(),
            ...     history=[Result("from history", ResultKind.VARIABLES)])

            >>> h.filter_by(kind={"MODEL"})   # doctest: +NORMALIZE_WHITESPACE
            History([Result(data='m1', kind=ResultKind.MODEL),
                                    Result(data='m2', kind=ResultKind.MODEL)])

            >>> h.filter_by(kind={ResultKind.OBSERVATION})  # doctest: +NORMALIZE_WHITESPACE
            History([Result(data='o1', kind=ResultKind.OBSERVATION),
                                    Result(data='o2', kind=ResultKind.OBSERVATION)])

            If we don't specify any filter criteria, we get the full history back:
            >>> h.filter_by()   # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            History([Result(data='from history', kind=ResultKind.VARIABLES),
                     Result(data=VariableCollection(...), kind=ResultKind.VARIABLES),
                     Result(data={'a': 'param'}, kind=ResultKind.PARAMS),
                     Result(data='c1', kind=ResultKind.CONDITION),
                     Result(data='c2', kind=ResultKind.CONDITION),
                     Result(data='o1', kind=ResultKind.OBSERVATION),
                     Result(data='o2', kind=ResultKind.OBSERVATION),
                     Result(data='m1', kind=ResultKind.MODEL),
                     Result(data='m2', kind=ResultKind.MODEL)])

        """
        if kind is None:
            return self
        else:
            kind_ = {ResultKind(s) for s in kind}
            filtered_history = _filter_history(self.data, kind_)
            new_object = History(history=filtered_history)
            return new_object


@dataclass(frozen=True)
class Result(SupportsDataKind):
    """
    Container class for data and variables.

    Examples:
        >>> Result()
        Result(data=None, kind=None)

        >>> Result("a")
        Result(data='a', kind=None)

        >>> Result(None, "MODEL")
        Result(data=None, kind=ResultKind.MODEL)

        >>> Result(data="b")
        Result(data='b', kind=None)

        >>> Result("c", "OBSERVATION")
        Result(data='c', kind=ResultKind.OBSERVATION)
    """

    data: Optional[Any] = None
    kind: Optional[ResultKind] = None

    def __post_init__(self):
        if isinstance(self.kind, str):
            object.__setattr__(self, "kind", ResultKind(self.kind))


def _init_result_list(
    variables: Optional[VariableCollection] = None,
    params: Optional[Dict] = None,
    conditions: Optional[Iterable[ArrayLike]] = None,
    observations: Optional[Iterable[ArrayLike]] = None,
    models: Optional[Iterable[BaseEstimator]] = None,
) -> List[Result]:
    """
    Initialize a list of Result objects

    Returns:

    Args:
        variables: a single datum to be marked as "variables"
        params: a single datum to be marked as "params"
        conditions: an iterable of data, each to be marked as "conditions"
        observations: an iterable of data, each to be marked as "observations"
        models: an iterable of data, each to be marked as "models"

    Examples:
        Empty input leads to an empty state:
        >>> _init_result_list()
        []

        ... or with values for any or all of the parameters:
        >>> from autora.variable import VariableCollection
        >>> _init_result_list(variables=VariableCollection()) # doctest: +ELLIPSIS
        [Result(data=VariableCollection(...), kind=ResultKind.VARIABLES)]

        >>> _init_result_list(params={"some": "params"})
        [Result(data={'some': 'params'}, kind=ResultKind.PARAMS)]

        >>> _init_result_list(conditions=["a condition"])
        [Result(data='a condition', kind=ResultKind.CONDITION)]

        >>> _init_result_list(observations=["an observation"])
        [Result(data='an observation', kind=ResultKind.OBSERVATION)]

        >>> from sklearn.linear_model import LinearRegression
        >>> _init_result_list(models=[LinearRegression()])
        [Result(data=LinearRegression(), kind=ResultKind.MODEL)]

        The input arguments are added to the data in the order `variables`,
        `params`, `conditions`, `observations`, `models`:
        >>> _init_result_list(variables=VariableCollection(),
        ...                  params={"some": "params"},
        ...                  conditions=["a condition"],
        ...                  observations=["an observation", "another observation"],
        ...                  models=[LinearRegression()],
        ... ) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        [Result(data=VariableCollection(...), kind=ResultKind.VARIABLES),
         Result(data={'some': 'params'}, kind=ResultKind.PARAMS),
         Result(data='a condition', kind=ResultKind.CONDITION),
         Result(data='an observation', kind=ResultKind.OBSERVATION),
         Result(data='another observation', kind=ResultKind.OBSERVATION),
         Result(data=LinearRegression(), kind=ResultKind.MODEL)]

    """
    data = []

    if variables is not None:
        data.append(Result(variables, ResultKind.VARIABLES))

    if params is not None:
        data.append(Result(params, ResultKind.PARAMS))

    for seq, kind in [
        (conditions, ResultKind.CONDITION),
        (observations, ResultKind.OBSERVATION),
        (models, ResultKind.MODEL),
    ]:
        if seq is not None:
            for i in seq:
                data.append(Result(i, kind=kind))

    return data


def _history_to_kind(history: Sequence[Result]) -> Snapshot:
    """
    Convert a sequence of results into a Snapshot instance:

    Examples:
        History might be empty
        >>> history_ = []
        >>> _history_to_kind(history_) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Snapshot(variables=VariableCollection(...), params={},
                        conditions=[], observations=[], models=[])

        ... or with values for any or all of the parameters:
        >>> history_ = _init_result_list(params={"some": "params"})
        >>> _history_to_kind(history_) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Snapshot(..., params={'some': 'params'}, ...)

        >>> history_ += _init_result_list(conditions=["a condition"])
        >>> _history_to_kind(history_) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Snapshot(..., params={'some': 'params'}, conditions=['a condition'], ...)

        >>> _history_to_kind(history_).params
        {'some': 'params'}

        >>> history_ += _init_result_list(observations=["an observation"])
        >>> _history_to_kind(history_) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Snapshot(..., params={'some': 'params'}, conditions=['a condition'],
                        observations=['an observation'], ...)

        >>> from sklearn.linear_model import LinearRegression
        >>> history_ = [Result(LinearRegression(), kind=ResultKind.MODEL)]
        >>> _history_to_kind(history_) # doctest: +ELLIPSIS
        Snapshot(..., models=[LinearRegression()])

        >>> from autora.variable import VariableCollection, IV
        >>> variables = VariableCollection(independent_variables=[IV(name="example")])
        >>> history_ = [Result(variables, kind=ResultKind.VARIABLES)]
        >>> _history_to_kind(history_) # doctest: +ELLIPSIS
        Snapshot(variables=VariableCollection(independent_variables=[IV(name='example', ...

        >>> history_ = [Result({'some': 'params'}, kind=ResultKind.PARAMS)]
        >>> _history_to_kind(history_) # doctest: +ELLIPSIS
        Snapshot(..., params={'some': 'params'}, ...)

    """
    namespace = Snapshot(
        variables=_get_last_data_with_default(
            history, kind={ResultKind.VARIABLES}, default=VariableCollection()
        ),
        params=_get_last_data_with_default(
            history, kind={ResultKind.PARAMS}, default={}
        ),
        observations=_list_data(
            _filter_history(history, kind={ResultKind.OBSERVATION})
        ),
        models=_list_data(_filter_history(history, kind={ResultKind.MODEL})),
        conditions=_list_data(_filter_history(history, kind={ResultKind.CONDITION})),
    )
    return namespace


def _list_data(data: Sequence[SupportsDataKind]):
    """
    Extract the `.data` attribute of each item in a sequence, and return as a list.

    Examples:
        >>> _list_data([])
        []

        >>> _list_data([Result("a"), Result("b")])
        ['a', 'b']
    """
    return list(r.data for r in data)


def _filter_history(data: Iterable[SupportsDataKind], kind: Set[ResultKind]):
    return filter(lambda r: r.kind in kind, data)


def _get_last(data: Sequence[SupportsDataKind], kind: Set[ResultKind]):
    results_new_to_old = reversed(data)
    last_of_kind = next(_filter_history(results_new_to_old, kind=kind))
    return last_of_kind


def _get_last_data_with_default(data: Sequence[SupportsDataKind], kind, default):
    try:
        result = _get_last(data, kind).data
    except StopIteration:
        result = default
    return result



import random
from typing import Iterable, List, Tuple

import numpy as np

from autora.utils.deprecation import deprecated_alias
from autora.variable import IV


def random_pool(
    ivs: List[IV], num_samples: int = 1, duplicates: bool = True
) -> Iterable:
    """
    Creates combinations from lists of discrete values using random selection.
    Args:
        ivs: List of independent variables
        n: Number of samples to sample
        duplicates: Boolean if duplicate value are allowed.

    """
    l_samples: List[Tuple] = []
    # Create list of pools of values sample from
    l_iv_values = []
    for iv in ivs:
        assert iv.allowed_values is not None, (
            f"gridsearch_pool only supports independent variables with discrete allowed values, "
            f"but allowed_values is None on {iv=} "
        )
        l_iv_values.append(iv.allowed_values)

    # Check to ensure infinite search won't occur if duplicates not allowed
    if not duplicates:
        l_pool_len = [len(set(s)) for s in l_iv_values]
        n_combinations = np.product(l_pool_len)
        try:
            assert num_samples <= n_combinations
        except AssertionError:
            raise AssertionError(
                f"Number to sample n({num_samples}) is larger than the number "
                f"of unique combinations({n_combinations})."
            )

    # Random sample from the pools until n is met
    while len(l_samples) < num_samples:
        l_samples.append(tuple(map(random.choice, l_iv_values)))
        if not duplicates:
            l_samples = [*set(l_samples)]

    return iter(l_samples)


random_pooler = deprecated_alias(random_pool, "random_pooler")
